# outer __init__.py

from .pandas.drop import drop
from .pandas.makes_up import makes_up
from .pandas.resemble import resemble
from .pandas.distort import distort
from .pandas.subset_by_levels import subset_by_levels
from .pandas.polynomializer import polynomializer
from .pandas.move_column_inplace import move_column_inplace

from .groups.fold import fold
from .groups.group_uniques import group_uniques
from .groups.group import group
from .groups.partition import partition

from .array.blend import blend
from .array.windowed_reverse import windowed_reverse
from .array.window import window
from .array.nan_stats import nan_stats, print_nan_stats

from .time.timestamps import Timestamps
from .time.timer import StepTimer

from .path.iopaths import IOPaths
from .path.mk_rm_dir import mk_dir, rm_dir

from .string.letter_strings import letter_strings
from .string.random_strings import random_alphanumeric

from .utils.messenger import Messenger, check_messenger, msg_if


def get_version():
    import importlib.metadata

    return importlib.metadata.version("utipy")


__version__ = get_version()
